/*
********************************************************************
**                                                                **
**	Tab Counter  -  Extension for Pale Moon                       **
**                                                                **
**	Last edit: September 10, 2018                                 **
**                                                                **
**	Release number: 1                                             **
**                                                                **
********************************************************************
*/
const {classes: Cc, interfaces: Ci, utils: Cu} = Components;
const PREF_BRANCH = "extensions.tabcounterpm.";
const PREFS = {
	fontColor: "", fontSize: "", fontFamily: "", fontWeight: "", backgrndColor: "#FFCC33", preLabel: "", postLabel: "",
	useFormatString: false, formatString: "%n", useGlobalCounts: false, noAppTabs: false, noTooltip: false, minWidth: 0
};

Cu.import("resource://gre/modules/XPCOMUtils.jsm");
Cu.import("resource://gre/modules/Services.jsm");

let global = this;
let addonData = null;
let disabled = true;

/* Includes a javascript file with loadSubScript.
*
* @param src (String)
* The url of a javascript file to include.
*/
(function(global) global.include = function include(src) {
	var o = {};
	Cu.import("resource://gre/modules/Services.jsm", o);
	var uri = o.Services.io.newURI(src, null, o.Services.io.newURI(__SCRIPT_URI_SPEC__, null, null));
	o.Services.scriptloader.loadSubScript(uri.spec, global);
})(this);

/* Imports a commonjs style javascript file with loadSubScrpt.
 * 
 * @param src (String)
 * The url of a javascript file.
 */
(function(global) {
	var modules = {};
	global.require = function require(src) {
		if(modules[src]) return modules[src];
		var scope = {require: global.require, exports: {}};
		var tools = {};
		Cu.import("resource://gre/modules/Services.jsm", tools);
		var baseURI = tools.Services.io.newURI(__SCRIPT_URI_SPEC__, null, null);
		try{
			var uri = tools.Services.io.newURI("packages/" + src + ".js", null, baseURI);
			tools.Services.scriptloader.loadSubScript(uri.spec, scope);
		} catch(e) {
			var uri = tools.Services.io.newURI(src, null, baseURI);
			tools.Services.scriptloader.loadSubScript(uri.spec, scope);
		};
		return modules[src] = scope.exports;
	};
})(this);

/**
 * Initialize default preferences specified in PREFS.
 */
function setDefaultPrefs() {
	let branch = Services.prefs.getDefaultBranch(PREF_BRANCH);
	for(let [key, val] in Iterator(PREFS)) {
		switch(typeof val) {
			case "boolean":
				branch.setBoolPref(key, val);
				break;
			case "number":
				if(Math.round(val) == val) {
					branch.setIntPref(key, val);
				};
				break;
			case "string":
				branch.setCharPref(key, val);
				break;
		};
	};
};

include("includes/buttons.js");

var {unload} = require("unload");
var {runOnLoad, runOnWindows, watchWindows} = require("window-utils");
var {loadOverlay, applyOverlay} = require("overlay");

function install(params, reason) {};
function uninstall(params, reason) {
	// Remove preferences when uninstalling.
	if(reason == ADDON_UNINSTALL) {
		Services.prefs.deleteBranch(PREF_BRANCH);
	};
};

function startup(params, reason) {
	disabled = false;
	addonData = params;

	if(reason == ADDON_INSTALL || reason == ADDON_ENABLE || reason == ADDON_UPGRADE || reason == ADDON_DOWNGRADE) {
	// Browser is already running in all these cases.
		Cu.import("chrome://tabcounter-pm/content/tbcGlobal.jsm", startup);
		startup.globalVal.getSetVal("tbcEnabled", "setT"); // set true.
		startup.globalVal.getSetVal("ssWinRest", "setT"); // set true.
	};

	// If installing, set default button position at the right edge of tabs toolbar, before the "List all tabs" button.
	if(reason == ADDON_INSTALL) {
		setDefaultPosition("tabcounterToolbar_item", "TabsToolbar", "alltabs-button");
	};

	// Load default preferences.
	setDefaultPrefs();

	// Check for presence & status of "Tab Groups" extension. Since this is
	// not a restartless extension, a single check on app startup is enough.
	Cu.import("resource://gre/modules/AddonManager.jsm", startup);
	startup.AddonManager.getAddonByID("firefox-tabgroups@mozilla.com", function(ftg) {
		if(ftg && ftg.isActive) {
			Cu.import("chrome://tabcounter-pm/content/tbcGlobal.jsm", startup);
			startup.globalVal.getSetVal("tabGroupsExt", "setT"); // set true.
			startup.AddonManager.getAddonByID("{1280606b-2510-4fe0-97ef-9b5a22eafe30}", function(sm) {
				if(sm && sm.isActive) { // "Session Manager" extension.
					startup.globalVal.getSetVal("sManagerExt", "setT"); // set true.
				};
			});
		};
	});

	// Read in window overlay.
	loadOverlay();

	// Register stylesheet.
	let sss = Cc["@mozilla.org/content/style-sheet-service;1"].getService(Ci.nsIStyleSheetService);
	let ios = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
	let uri = ios.newURI("chrome://tabcounterpm/content/tbcItem.css", null, null);
	sss.loadAndRegisterSheet(uri, sss.USER_SHEET);

	// Watch for existing and newly opened windows.
	watchWindows(applyOverlay, "navigator:browser");

	// Add unload functions
	unload(function() {
		// If options window open, close it.
		let win = Services.wm.getMostRecentWindow("TabCounterPm:Options");
		if(win) {
			win.close();
			Services.strings.flushBundles();
		};
	});
};

function shutdown(params, reason) {
	disabled = true;

	// Don't both unloading if shutting down.
	if(reason == APP_SHUTDOWN) {
		return;
	};

	Cu.unload("chrome://tabcounter-pm/content/tbcGlobal.jsm");

	if(reason == ADDON_UPGRADE || reason == ADDON_DOWNGRADE) {
		// HACK WARNING: The Addon Manager does not properly clear all addon related caches on update;
		//				 in order to fully update images and locales, their caches need clearing here.
		// https://developer.mozilla.org/en-US/docs/Archive/Add-ons/How_to_convert_an_overlay_extension_to_restartless
		Services.obs.notifyObservers(null, "chrome-flush-caches", null);
	};

	//  Unregister stylesheet.
	let sss = Cc["@mozilla.org/content/style-sheet-service;1"].getService(Ci.nsIStyleSheetService);
	let ios = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
	let uri = ios.newURI("chrome://tabcounterpm/content/tbcItem.css", null, null);
	if(sss.sheetRegistered(uri, sss.USER_SHEET)) {
		sss.unregisterSheet(uri, sss.USER_SHEET);
	};
	unload();
	addonData = null;
};
