/*
********************************************************************
**                                                                **
**	Tab Counter  -  Extension for Pale Moon                       **
**                                                                **
**	Last edit: September 13, 2018                                 **
**                                                                **
**	Release number: 1                                             **
**                                                                **
********************************************************************
*/
var gTabCounterPm = {
	lastHiddenTabsCount: 0,
	ssBrStateRest: false, // "sessionstore-browser-state-restored".

	initialize: function() { // Do stuff on browser startup or when new browser window opens.
		if(!gTabCounterPm.globalVal.getSetVal("ssWinRest", "getV")) { // get value.
		// Workaround to prevent multiple runs of "tabEventHandler()" or "updateToolbarItem()" due to
		// burst of events if there are many tabs to be restored on browser startup. After this
		// notification also, "browsers.length" & "selectedTab._tPos" we use are always correct.
			Services.obs.addObserver(gTabCounterPm, "sessionstore-windows-restored", false);
		} else { // New or restored browser window opens or we are installed/enabled/upgraded/downgraded.
			var start2Count = function gT_start2Count(aEvent) {
				if(gTabCounterPm.globalVal.getSetVal("tabGroupsExt", "getV")) {
					var ss = Cc["@mozilla.org/browser/sessionstore;1"].getService(Ci.nsISessionStore);
					if(!this.ssBrStateRest &&
						((ss.canRestoreLastSession && !PrivateBrowsingUtils.isWindowPrivate(window)) ||
						gTabCounterPm.globalVal.getSetVal("sManagerExt", "getV"))) { // "Session Manager" allows
						// session restorations in private windows, while should not..
						Services.obs.addObserver(gTabCounterPm, "sessionstore-browser-state-restored", false);
						this.ssBrStateRest = true;
					};
					this.lastHiddenTabsCount = gBrowser.browsers.length - gBrowser.visibleTabs.length;
				};
				if(!aEvent || (aEvent.type == "DOMWindowCreated")) { // !aEvent -> we are installed/enabled/upgraded/downgraded.
					if(aEvent && (aEvent.type == "DOMWindowCreated")) { // On "Restore All Windows", only the
					// last must be allowed to run its iterator to update the global counts in rest windows.
						if(gTabCounterPm.globalVal.getSetVal("domReadyWin", "getV") !==
							gTabCounterPm.globalVal.getSetVal("openWindows", "getV")) { // get value.
							let sum = gTabCounterPm.globalVal.getSetVal("domReadyWin", "getV") + 1;
							gTabCounterPm.globalVal.getSetVal("domReadyWin", "setV", sum);
							if(gTabCounterPm.globalVal.getSetVal("domReadyWin", "getV") !==
								gTabCounterPm.globalVal.getSetVal("openWindows", "getV")) { // get value.
								gTabCounterPm.globalVal.getSetVal("canUpdateOtherWin", "setF"); // set false.
							} else {
								gTabCounterPm.globalVal.getSetVal("canUpdateOtherWin", "setT"); // set true.
							};
						};
					};
					this.updateToolbarItem();
					this.addLis();
				return;
				};
				if(aEvent.type == "SSWindowStateReady") { // Session restore in process, not on browser startup.
					if(!this.onSessionRestore(aEvent)) { // "onSessionRestore()" is handled as a callback here.
						this.updateToolbarItem();
						this.addLis();
					} else {
						this.addLis();
					};
				};
			}.bind(this);

			if(gTabCounterPm.globalVal.getSetVal("tbcEnabled", "getV")) { // get value - We are installed/enabled/upgraded/downgraded.
				start2Count();
			} else { // New or restored browser window opens.
				var winOpenEventHandler = (aEvent) => {
					aEvent.type == "DOMWindowCreated" ?
					window.removeEventListener("DOMWindowCreated", winOpenEventHandler, false) :
					window.removeEventListener("SSWindowStateReady", winOpenEventHandler, false);
					window.setTimeout(() => {
						start2Count(aEvent);
					}, 0);
				};
				if(gTabCounterPm.globalVal.getSetVal("browserIsBusy", "getV")) { // get value - Session restore in process, not on browser startup.
					window.addEventListener("SSWindowStateReady", winOpenEventHandler, false);
				} else {
					window.addEventListener("DOMWindowCreated", winOpenEventHandler, false); // This event is
					// fired late enough, so our count sources get finalized. This is what we want, if the newly
					// opened window is a restored window with more than 1 tab or many tabs or even tab groups.
				};
			};
		};
	},

	uninitialize: function gT_uninitialize(winClose) { // Do stuff when browser window closes or we get be disabled.
		window.removeEventListener("SSWindowStateBusy", this.onSessionRestore, false);
		var container = gBrowser.tabContainer;
		container.removeEventListener("TabOpen", this.proxy, false);
		container.removeEventListener("TabClose", this.proxy, false);
		container.removeEventListener("TabMove", this.proxy, false);
		container.removeEventListener("TabSelect", this.proxy, false);
		container.removeEventListener("TabPinned", this.proxy, false);
		container.removeEventListener("TabUnpinned", this.proxy, false);
		if(gTabCounterPm.globalVal.getSetVal("tabGroupsExt", "getV")) { // get value.
			window.removeEventListener("tabviewhidden", this.proxy, false);
		};
		if(this.ssBrStateRest) {
			Services.obs.removeObserver(gTabCounterPm, "sessionstore-browser-state-restored");
		};
		if(winClose) { // No need and will not run on extension's disable/uninstall.
			let subtractWin = gTabCounterPm.globalVal.getSetVal("openWindows", "getV") - 1;
			gTabCounterPm.globalVal.getSetVal("openWindows", "setV", subtractWin); // set value.
			let restOpenWin = gTabCounterPm.globalVal.getSetVal("openWindows", "getV");
			gTabCounterPm.globalVal.getSetVal("domReadyWin", "setV", restOpenWin); // set value. - "domReadyWin": info in "tbcGlobal.jsm".
			if((gTabCounterPm.globalVal.getSetVal("openWindows", "getV") >= 1) &&
				this.getPref("extensions.tabcounterpm.useGlobalCounts", Ci.nsIPrefBranch.PREF_BOOL)) {
				let currentTotalCount_S = gTabCounterPm.globalVal.getSetVal("totalCount_S", "getV"); // get value.
				gTabCounterPm.globalVal.getSetVal("totalCount_S", "setV", currentTotalCount_S - this.lastTotalCountCW); // set value.
				let currentVisCount_S = gTabCounterPm.globalVal.getSetVal("visCount_S", "getV"); // get value.
				gTabCounterPm.globalVal.getSetVal("visCount_S", "setV", currentVisCount_S - this.lastVisCountCW); // set value.
				let openBrWinEnum = Services.wm.getEnumerator("navigator:browser");
				while(openBrWinEnum.hasMoreElements()) {
					let brWin = openBrWinEnum.getNext();
					if((brWin.window !== window.window) && brWin.gTabCounterPm) {
						brWin.gTabCounterPm.updateToolbarItem("inProcess"); // Update the "per session" counts in rest windows.
						// Normally the argument is not needed here, added just in case something goes wrong, to prevent
						// multiple/infinite iterations in "updateToolbarItem()".
					};
				};
			};
		};
		gTabCounterPm = null;
	},

	addLis: function() {
		window.addEventListener("SSWindowStateBusy", this.onSessionRestore, false);
		var container = gBrowser.tabContainer;
		container.addEventListener("TabOpen", this.proxy, false);
		container.addEventListener("TabClose", this.proxy, false);
		container.addEventListener("TabMove", this.proxy, false);
		container.addEventListener("TabSelect", this.proxy, false);
		container.addEventListener("TabPinned", this.proxy, false);
		container.addEventListener("TabUnpinned", this.proxy, false);
		if(gTabCounterPm.globalVal.getSetVal("tabGroupsExt", "getV")) { // get value.
			window.addEventListener("tabviewhidden", this.proxy, false);
		};
	},

	proxy: function(aEvent) { // Assigning the desired "this" in tabEventHandler() - (object XULElement/ChromeWindow -> object gTabCounterPm).
		gTabCounterPm.tabEventHandler(aEvent);
	},

	observe: function gT_observe(aSubject, aTopic, aData) {
		switch(aTopic) {
			case "sessionstore-windows-restored": // Browser startup.
				gTabCounterPm.globalVal.getSetVal("ssWinRest", "setT"); // set true.
				Services.obs.removeObserver(gTabCounterPm, "sessionstore-windows-restored");
				if(gTabCounterPm.globalVal.getSetVal("tabGroupsExt", "getV")) {
					var ss = Cc["@mozilla.org/browser/sessionstore;1"].getService(Ci.nsISessionStore);
					if(ss.canRestoreLastSession || gTabCounterPm.globalVal.getSetVal("sManagerExt", "getV")) {
						Services.obs.addObserver(gTabCounterPm, "sessionstore-browser-state-restored", false);
						this.ssBrStateRest = true;
					};
				};
				
				// Session restoration on browser startup: details in "window-utils.js", "runOnLoad()".
				if(gTabCounterPm.globalVal.getSetVal("restoredWindows", "getV") == 0) { // get value.
					let restoredWindows = 0;
					let restoredWinEnum = Services.wm.getEnumerator("navigator:browser");
					while(restoredWinEnum.hasMoreElements()) {
						let restBrWin = restoredWinEnum.getNext();
						if(restBrWin.window) {
							restoredWindows++;
						};
					};
					if(restoredWindows > 1) {
						gTabCounterPm.globalVal.getSetVal("restoredWindows", "setV", restoredWindows); // set value.
					} else {
						gTabCounterPm.globalVal.getSetVal("canUpdateOtherWin", "setT"); // set true.
					};
					// "domReadyWin": info in "tbcGlobal.jsm".
					gTabCounterPm.globalVal.getSetVal("domReadyWin", "setV", restoredWindows); // set value.
				};

				this.lastHiddenTabsCount = gBrowser.browsers.length - gBrowser.visibleTabs.length;
				this.updateToolbarItem(); // Update tab count of current state.
				this.addLis();
				break;
			case "sessionstore-browser-state-restored":
				if(!gTabCounterPm.globalVal.getSetVal("sManagerExt", "getV")) { // get value.
					Services.obs.removeObserver(gTabCounterPm, "sessionstore-browser-state-restored");
					this.ssBrStateRest = false;
				};
				// Necessary update in case there are hidden tabs (Tab Groups), effective only if the selected tab of
				// the restored session is the first (No 1 in order). Otherwise, we are covered by the "TabSelect" event.
				this.lastHiddenTabsCount = gBrowser.browsers.length - gBrowser.visibleTabs.length;
				break;
		};
	},

	tabEvents: [], // In sequential events, holds the current[0] and the previous[1] event, until "updateToolbarItem()" is run.
	timeoutId: undefined, // Used in event time frame.
	tabEventHandler: function(aEvent) { // In case of multiple calls due to burst of events of any type of those
	// we are listening to, we proceed only on the last call, which also corresponds to the final tabs status,
	// to prevent multiple runs of "updateToolbarItem()". Also, the small delay of our event time frame addresses the
	// issue where "visibleTabs.length" is not correct immediately after "tabviewhidden" event or after switching
	// groups with "CTRL+`" keyboard shortcut.
		var proceedPendingAction = function gT_proceedPendingAction(aEvent) {
			if(gTabCounterPm.globalVal.getSetVal("tabGroupsExt", "getV")) { // get value.
				if(aEvent) {
					let gTabViewDeck = window.document.getElementById("tab-view-deck");
					let gTabViewFrame = window.document.getElementById("tab-view");
					if(((aEvent.type == "TabSelect") && (this.tabEvents[1] !== "TabClose")) || // On switching groups with "CTRL+`".
						((aEvent.type == "tabviewhidden") && ((this.tabEvents[1] == "TabSelect") || (this.tabEvents[1] == "TabClose") || (this.tabEvents[1] == "TabMove") || !this.tabEvents[1])) ||
						((aEvent.type == "TabClose") && (gTabViewDeck.selectedPanel == gTabViewFrame)) || // On discarding tab group or closing tabs within tabview mode.
						((aEvent.type == "TabSelect") && (this.tabEvents[1] == "TabClose") && (gTabViewDeck.selectedPanel == gTabViewFrame))) { // On closing the last tab in a group within tabview mode.
						this.lastHiddenTabsCount = gBrowser.browsers.length - gBrowser.visibleTabs.length; // Necessary
						// update in case of switching groups, interchanging tabs between tab groups, discarding tab groups 
						// within or on exiting tabview mode, but strictly undesirable when closing the selected tab
						// ("TabSelect" event will occur), or multiple tabs including or not the selected, or just one tab,
						// because "browsers.length" is not instantly correct after "TabClose" event.
						// By the rules above, we try to distinct what's the case and accordingly we update or not the "lastHiddenTabsCount".
					};
				};
			};
			this.timeoutId = undefined;
			this.updateToolbarItem();
		}.bind(this);

		var delay = 50; // Sufficient time frame for sequential events.
		var setPendingAction = function(aEvent) {
			if(typeof this.timeoutId === "number") {
				window.clearTimeout(this.timeoutId);
				this.timeoutId = undefined;
			};
			if(aEvent && gTabCounterPm.globalVal.getSetVal("tabGroupsExt", "getV")) { // get value.
				let gTabViewDeck = window.document.getElementById("tab-view-deck");
				let gTabViewFrame = window.document.getElementById("tab-view");
				if((aEvent.type == "TabSelect") && (this.tabEvents[1] == "TabOpen") && (gTabViewDeck.selectedPanel == gTabViewFrame)) {
					delay = 350; // Selecting an empty group or double-clicking on free space in tabview mode, we have this
					// event sequence: "TabOpen" -> "TabSelect" -> "tabviewhidden". In that case however, the "tabviewhidden"
					// event takes considerably long to occur. Strange thing..
					// Increasing the time frame, we just prevent "updateToolbarItem()" to run two times.
				};
			};
			this.timeoutId = window.setTimeout(() => {
				proceedPendingAction(aEvent);
			}, delay);
		}.bind(this);

		if(aEvent) {
			this.tabEvents.unshift(aEvent.type);
			if(this.tabEvents.length > 2) {
				this.tabEvents.pop();
			};
			switch(aEvent.type) {
				case "TabOpen":
				case "TabMove":
				case "TabPinned":
				case "TabUnpinned":
					setPendingAction();
					break;
				case "TabClose":
				case "TabSelect":
				case "tabviewhidden":
					delay = 100; // A little bigger time frame is required in these cases.
					setPendingAction(aEvent);
					break;
			};
		};
	},

	onCustomize: function() { // Called by "runOnLoad()" in "window-utils.js". - This method does not work on Firefox 29+.
	// When our toolbar item is moved in customization palette, if its label is too long to fit in the available limited space, the
	// result will be a very ugly appearance in palette. We could check the width of the label but it's better to have always a default
	// appearance within customization palette. So, when our item is moved in palette, we strip it out of any custom attributes.
		var tbcTlbrItem = window.document.getElementById("tabcounterToolbar_item");
		var paletteBox = null;
		var dropLis = false;
		function onDropInPalette() { // Our toolbar item must be stripped out in both places where coexist during the drag & drop process.
			if(paletteBox.hasChildNodes()) {
				if(paletteBox.lastChild.childNodes[0].id == "tabcounterToolbar_item") { // A copy of our toolbar item just dropped in palette.
					// The original item is still on toolbar, awaiting its removal by "onPaletteDrop()" (Browser's "customizeToolbar.js").
					tbcTlbrItem.childNodes[0].setAttribute("value", "?");
					tbcTlbrItem.childNodes[0].removeAttribute("style");
					tbcTlbrItem.style.backgroundColor = "#FFCC33";
					tbcTlbrItem.style.removeProperty("min-width");
					// The copy of our toolbar item in customization palette.
					paletteBox.lastChild.childNodes[0].childNodes[0].setAttribute("value", "?");
					paletteBox.lastChild.childNodes[0].childNodes[0].removeAttribute("style")
					paletteBox.lastChild.childNodes[0].style.backgroundColor = "#FFCC33";
					paletteBox.lastChild.childNodes[0].style.removeProperty("min-width");
					// The "ondrop" listener is not needed any more.
					paletteBox.removeEventListener("drop", onDropInPalette, false);
					dropLis = false;
				};
			};
		};
		if(tbcTlbrItem) { // Our toolbar item is on a toolbar.
			let allWin = Services.wm.getEnumerator(null); // There is no "windowtype" in customization window..
			while(allWin.hasMoreElements()) {
				let win = allWin.getNext();
				if((win.document.documentElement.getAttribute("id") == "CustomizeToolbarWindow") &&
					(win.opener == Services.wm.getMostRecentWindow("navigator:browser"))) { // Pale Moon allows more
					// than 1 customization window when there are normal & private browser windows, while should not..
					paletteBox = win.document.getElementById("palette-box");
					break;
				};
			};
			paletteBox.addEventListener("drop", onDropInPalette, false);
			dropLis = true;
		};
		function afterCustMode() {
			window.removeEventListener("aftercustomization", afterCustMode, false);
			if(dropLis) { // Our toolbar item did not moved in palette during customization.
				paletteBox.removeEventListener("drop", onDropInPalette, false);
				dropLis = false;
			};
			paletteBox = null;
			tbcTlbrItem = null;
			if(window.document.getElementById("tabcounterToolbar_item") &&
				!window.document.getElementById("tabcounterToolbar_item").childNodes[0].hasAttribute("style")) {
				gTabCounterPm.styleApplied = false;
				gTabCounterPm.updateToolbarItem();
			};
		};
		window.addEventListener("aftercustomization", afterCustMode, false);
	},

	expandDisplayStringFormats: function(formatString, visCountCW, visCurrentTab, totalCountCW, totalCurrentTab, pinned, totalCount_S, visCount_S) {
		var expandedString = formatString.replace("%" + this._string("formatStringToolTip.after.key"), pinned ? totalCountCW : (totalCountCW - totalCurrentTab));
		expandedString = expandedString.replace("%" + this._string("formatStringToolTip.before.key"), pinned ? "0" : (totalCurrentTab - 1));
		expandedString = expandedString.replace("%" + this._string("formatStringToolTip.current.key"), pinned ? "0" : totalCurrentTab);
		expandedString = expandedString.replace("%" + this._string("formatStringToolTip.number.key"), totalCountCW);
		expandedString = expandedString.replace("%" + this._string("formatStringToolTip.number_s.key"), totalCount_S);
		if(gTabCounterPm.globalVal.getSetVal("tabGroupsExt", "getV")) { // get value.)
			expandedString = expandedString.replace("%" + this._string("formatStringToolTip.visible_after.key"), pinned ? visCountCW : (visCountCW - visCurrentTab));
			expandedString = expandedString.replace("%" + this._string("formatStringToolTip.visible_before.key"), pinned ? "0" : (visCurrentTab - 1));
			expandedString = expandedString.replace("%" + this._string("formatStringToolTip.visible_current.key"), pinned ? "0" : visCurrentTab);
			expandedString = expandedString.replace("%" + this._string("formatStringToolTip.visible_number.key"), visCountCW);
			expandedString = expandedString.replace("%" + this._string("formatStringToolTip.visible_number_s.key"), visCount_S);
		};
		return expandedString;
	},

	// The offset is used to correct when ignoring app tabs.
	findTabIndex: function(aTab, offSet) { // aTab = gBrowser.selectedTab, offSet = pinnedTabCountCW.
		var totalCurrent = null;
		// Find total current tab.
		// For Firefox/Pale Moon, the tab index is stored in _tPos. If that doesn't exist, do a search.
		if(typeof aTab._tPos != "undefined") {
			totalCurrent = aTab._tPos;
		} else {
			// Check each tab of this browser instance.
			for(var index = 0; index < aTab.parentNode.childNodes.length; index++) {
				if(aTab == aTab.parentNode.childNodes[index]) {
					totalCurrent = index;
					break;
				};
			};
		};

		// Find visible current tab if exists, otherwise use total current.
		// If visible tabs exists and not the same length as browsers, then there are hidden tabs so calculate tab "index" manually.
		var visibleCurrent = null;
		if(gBrowser.visibleTabs && (gBrowser.visibleTabs.length !== (gBrowser.visibleTabs.length + this.lastHiddenTabsCount))) {
			// Check each visible tab of this browser instance
			for(var index = 0; index < gBrowser.visibleTabs.length; index++) {
				if(aTab == gBrowser.visibleTabs[index]) {
					visibleCurrent = index;
					break;
				};
			};
		} else {
			visibleCurrent = totalCurrent;
		};
		return [visibleCurrent + 1 - offSet, totalCurrent + 1 - offSet]; // "+ 1" because the count starts from 0,
	},																	 // so, the first tab will have number 1.

	lastVisCountCW: 0, // Current window.
	lastTotalCountCW: 0, // Current window.
	storedTooltip: null,
	checkLoops: 0,
	checkLoopsDelay: 120,
	styleApplied: false,
	updateToolbarItem: function gT_updateToolbarItem(aParam) {
		// If not counting app (pinned) tabs, then get the pinned tab count.
		var pinnedTabCountCW = this.getPref("extensions.tabcounterpm.noAppTabs", Ci.nsIPrefBranch.PREF_BOOL) ? gBrowser._numPinnedTabs : 0;

		// For some reasons, after closing a tab or multiple tabs (e.g. with "Close Tabs to the Right"), "browsers.length" becomes correct
		// after an indefinable delay. The same applies also for "tabs.length" and "tabContainer.childNodes.length", which could be used as
		// a count source of all tabs. So, we get instantly the count of all tabs by using "lastHiddenTabsCount", which is already calculated
		// when "browsers.length" was correct. Without Tab Groups, "lastHiddenTabsCount" should be always 0.
		var allTabsCW = gBrowser.visibleTabs.length + this.lastHiddenTabsCount; // All tabs in current window.

		// Before we proceed and call "findTabIndex()", it is important for "selectedTab._tPos" to be correct. Based on it we calculate also the
		// "before" & "after" numbers. However, when closing a tab or tabs, "_tPos" becomes correct after an indefinable delay (The same applies
		// for "tabContainer.selectedIndex"). Fortunately, it gets updated (almost?) the same time with "browsers.length". So, here we check
		// (& recheck if/as needed) "browsers.length" and proceed when it becomes correct.
		if(gBrowser.browsers.length !== allTabsCW && this.checkLoops < 15) { // Usually takes ~150-250ms for "browsers.length" to became correct.
			this.checkLoops++; // Prevent endless loops under unpredictable conditions..
			window.setTimeout(() => {
				this.updateToolbarItem();
			}, this.checkLoopsDelay);
			if(this.checkLoopsDelay > 26) { // Limit smallest time frame to 26ms (fourth+ loop), if loops > 4 (unlikely).
				this.checkLoopsDelay = Math.round(this.checkLoopsDelay - this.checkLoopsDelay * .4); // Subtract 40% in each loop.
				// Initial time frame 120ms, next 72ms, 43ms, limited to 26ms (fourth+ loop).
			};
			return;
		} else if(this.checkLoops == 15) { // Last resort..
			this.checkLoops = 0;
			console.log("Tab Counter could not count the tabs correctly in time. Last resort was to try again after a delay.");
			window.setTimeout(() => { // We'll get here if "lastHiddenTabsCount" is not correct.
				this.lastHiddenTabsCount = gBrowser.browsers.length - gBrowser.visibleTabs.length;
				this.updateToolbarItem();
			}, 3000);
			return;
		};

		var visCountCW = gBrowser.visibleTabs.length - pinnedTabCountCW; // Current window.
		var totalCountCW = gTabCounterPm.globalVal.getSetVal("tabGroupsExt", "getV") ? allTabsCW - pinnedTabCountCW : visCountCW; // Current window.
		if(this.getPref("extensions.tabcounterpm.useGlobalCounts", Ci.nsIPrefBranch.PREF_BOOL)) { // Our method we use to count global numbers
		// (tabs in all windows): Each window (initially the first one) in which a tab event occurs, updates the global counts in a central
		// point, to be directly available for all existing and new windows and windows which getting to be closed, updates itself and finally
		// updates all other windows by means of triggering them to get the global counts from the central point and update themselves. It's a
		// much lighter process than counting all tabs in all windows at every tab event. However, although it seems to work, this method maybe
		// is more prone to be unreliable, due to unpredictable reasons. If something goes wrong and the global counts become incorrect, they
		// will not get self corrected by continue using the browser. It will need a browser restart or uncheck "Use advanced text formatting" in
		// our pref window, click "Apply", check "Use advanced text formatting", click "Apply/OK". By the time passing we'll see..
			if(gTabCounterPm.globalVal.getSetVal("visCount_S", "getV") == 0) { // First browser window or we are installed/enabled/upgraded/downgraded.
				gTabCounterPm.globalVal.getSetVal("visCount_S", "setV", visCountCW); // set value.
				this.lastVisCountCW = visCountCW;
			};
			// When "visCountCW" is changed, we update the global value of visible tabs number in all windows.
			if(visCountCW !== this.lastVisCountCW) { // It is true also when a new window opens, where
				let diffV = 0;						 // "lastVisCountCW" is 0.
				let currentVisCount_S = gTabCounterPm.globalVal.getSetVal("visCount_S", "getV"); // get value.
				if(visCountCW > this.lastVisCountCW) {
					diffV = visCountCW - this.lastVisCountCW;
					gTabCounterPm.globalVal.getSetVal("visCount_S", "setV", currentVisCount_S + diffV); // set value.
				} else if(visCountCW < this.lastVisCountCW) {
					diffV = this.lastVisCountCW - visCountCW;
					gTabCounterPm.globalVal.getSetVal("visCount_S", "setV", currentVisCount_S - diffV); // set value.
				};
			};
			if(gTabCounterPm.globalVal.getSetVal("totalCount_S", "getV") == 0) { // First browser window or we are installed/enabled/upgraded/downgraded.
				gTabCounterPm.globalVal.getSetVal("totalCount_S", "setV", totalCountCW); // set value.
				this.lastTotalCountCW = totalCountCW;
			};
			// As above, when "totalCountCW" is changed, we update the global value of total tabs number in all windows.
			if(totalCountCW !== this.lastTotalCountCW) { // It is true also when a new window opens, where
				let diffT = 0;							 // "lastTotalCountCW" is 0.
				let currentTotalCount_S = gTabCounterPm.globalVal.getSetVal("totalCount_S", "getV"); // get value.
				if(totalCountCW > this.lastTotalCountCW) {
					diffT = totalCountCW - this.lastTotalCountCW;
					gTabCounterPm.globalVal.getSetVal("totalCount_S", "setV", currentTotalCount_S + diffT); // set value.
				} else if(totalCountCW < this.lastTotalCountCW) {
					diffT = this.lastTotalCountCW - totalCountCW;
					gTabCounterPm.globalVal.getSetVal("totalCount_S", "setV", currentTotalCount_S - diffT); // set value.
				};
			};
		};
		var visCount_S = gTabCounterPm.globalVal.getSetVal("visCount_S", "getV"); // get value.
		var totalCount_S = gTabCounterPm.globalVal.getSetVal("totalCount_S", "getV"); // get value.

		var useFormatString = this.getPref("extensions.tabcounterpm.useFormatString", Ci.nsIPrefBranch.PREF_BOOL);
		var displayString = "";
		if(!useFormatString) {
			let preText = this.getPref("extensions.tabcounterpm.preLabel", Ci.nsIPrefBranch.PREF_STRING);
			let postText = this.getPref("extensions.tabcounterpm.postLabel", Ci.nsIPrefBranch.PREF_STRING);
			preText = preText ? preText + " " : "";
			postText = postText ? " " + postText : "";
			displayString = preText + visCountCW + postText;
		} else {
			var formatString = this.getPref("extensions.tabcounterpm.formatString", Ci.nsIPrefBranch.PREF_STRING);
			var currentTab = this.findTabIndex(gBrowser.selectedTab, pinnedTabCountCW);
			// dump(currentTab + "\n");
			if(currentTab[0] > visCountCW) {
				currentTab[0] = visCountCW;
			};
			if(currentTab[1] > totalCountCW) {
				currentTab[1] = totalCountCW;
			};
			// If a pinned tab is selected and not countning them don't use current values.
			displayString = this.expandDisplayStringFormats(formatString, visCountCW, currentTab[0], totalCountCW, currentTab[1], ((pinnedTabCountCW > 0) && gBrowser.selectedTab.getAttribute("pinned")), totalCount_S, visCount_S);
		};

		var tbcToolbarItem = document.getElementById("tabcounterToolbar_item");
		if(tbcToolbarItem) { // We check for and handle our toolbar item after we finish countings, due to a limitation of Pale Moon which affects
		// our global counts: If there are more than 1 window and our toolbar item is in customization palette, if it is placed on a toolbar,
		// Pale Moon does not update automatically the other open windows. Until a browser restart, our global counts would not be correct.
			var tbcToolbarItemLabel = document.getElementById("tabcounterDisplay_label");

			// Tricky way of getting "Open tabs here: Open windows: " locale text without requiring stringbundle - gets all text but numbers.
			if(this.getPref("extensions.tabcounterpm.noTooltip", Ci.nsIPrefBranch.PREF_BOOL)) {
				if(tbcToolbarItem.hasAttribute("tooltiptext")) { // Necessary check: In case "Disable tooltip" is checked and then "Apply" is
					// clicked and then "OK" is clicked, this function runs twice, but the second time "tbcToolbarItem" doesn't have "tooltiptext"
					// attribute, so "storedTooltip" will be flushed (null). This will happen also after any tab event which calls this function.
					// After that, when "Disable tooltip" is unchecked, it is impossible to restore the "Open tabs here: Open windows: " text,
					// leaving the re-enabled tooltip to display only numbers.
					this.storedTooltip = tbcToolbarItem.getAttribute("tooltiptext").match(/[^\d]+/g).join("");
					tbcToolbarItem.removeAttribute("tooltiptext");
				};
			} else {
				var strArray = (this.storedTooltip || tbcToolbarItem.getAttribute("tooltiptext").match(/[^\d]+/g).join("")).replace(/\n/, "").split(/(:\s)/, 4); // Returns [ "Open tabs here", ": ", "Open windows", ": " ].
				// var strArray = (this.storedTooltip || tbcToolbarItem.getAttribute("tooltiptext").match(/[^\d]+/g).join("")).replace(/\n/, "").split(/:\s/, 2); // Works the same.
				var str1 = strArray.slice(0, 2).join("") + visCountCW;
				var str2 = strArray.slice(2, 4).join("") + gTabCounterPm.globalVal.getSetVal("openWindows", "getV");
				// var str1 = strArray[0] + ": " + visCountCW; // Works the same.
				// var str2 = strArray[1] + ": " + gTabCounterPm.globalVal.getSetVal("openWindows", "getV"); // Works the same.
				tbcToolbarItem.setAttribute("tooltiptext", str1 + "\n" + str2);
				this.storedTooltip = null;
			};

			if(tbcToolbarItemLabel) {
				tbcToolbarItemLabel.setAttribute("value", displayString);
				if(!this.styleApplied) { // Apply style in our toolbar item only once per window and/or every time our options window is used.
					tbcToolbarItemLabel.style.color = this.getPref("extensions.tabcounterpm.fontColor", Ci.nsIPrefBranch.PREF_STRING);
					tbcToolbarItemLabel.style.fontSize = this.getPref("extensions.tabcounterpm.fontSize", Ci.nsIPrefBranch.PREF_STRING);
					tbcToolbarItemLabel.style.fontFamily = this.getPref("extensions.tabcounterpm.fontFamily", Ci.nsIPrefBranch.PREF_STRING);
					tbcToolbarItemLabel.style.fontWeight = this.getPref("extensions.tabcounterpm.fontWeight", Ci.nsIPrefBranch.PREF_STRING);
					tbcToolbarItemLabel.style.margin = "0 3px";
					tbcToolbarItem.style.backgroundColor = this.getPref("extensions.tabcounterpm.backgrndColor", Ci.nsIPrefBranch.PREF_STRING);
					tbcToolbarItem.style.minWidth = this.getPref("extensions.tabcounterpm.minWidth", Ci.nsIPrefBranch.PREF_INT) + "px";
					this.styleApplied = true;
				};
			};
		};
		window.setTimeout(() => {
			this.lastHiddenTabsCount = gBrowser.browsers.length - gBrowser.visibleTabs.length;
		}, 500);
		this.checkLoops = 0;
		this.checkLoopsDelay = 120;
		this.tabEvents = [];
		if(aParam == "inProcess") {
			this.lastVisCountCW = visCountCW;
			this.lastTotalCountCW = totalCountCW;
			return;
		} else if(aParam == "allRestoredWinReady") { // Trick to fool the following iterator, to make it run,
			this.lastVisCountCW = 0;				 // avoiding a far more complicated method..
			this.lastTotalCountCW = 0;
		};
		let useGlobalCounts = this.getPref("extensions.tabcounterpm.useGlobalCounts", Ci.nsIPrefBranch.PREF_BOOL);
		if(useGlobalCounts && (gTabCounterPm.globalVal.getSetVal("openWindows", "getV") > 1) &&
			gTabCounterPm.globalVal.getSetVal("canUpdateOtherWin", "getV") &&
			((totalCountCW !== this.lastTotalCountCW) || (visCountCW !== this.lastVisCountCW))) {
			let openBrWinEnum = Services.wm.getEnumerator("navigator:browser");
			while(openBrWinEnum.hasMoreElements()) {
				let brWin = openBrWinEnum.getNext();
				if((brWin.window !== window.window) && brWin.gTabCounterPm) {
					brWin.gTabCounterPm.updateToolbarItem("inProcess"); // Update the "per session" counts in other windows.
				};
			};
		};
		this.lastVisCountCW = visCountCW;
		this.lastTotalCountCW = totalCountCW;
	},

	onSessionRestore: function(aEvent) { // Not on browser startup!
	// On session restoration, we can know the number of restored windows before the last "SSWindowStateReady"
	// event. So, we compare them with the number of "State Ready Windows" to find which is the last, to
	// enable/allow it to update the global counts in other windows.
	// This method does not work on Firefox, due to different timings in events/obs notifications we use.
		let rootReturn = false;
		let restoredWindows = 0;
		function checkBrowserReadiness() {
			if(gTabCounterPm.globalVal.getSetVal("stateReadyWin", "getV") ==
				gTabCounterPm.globalVal.getSetVal("restoredWindows", "getV")) { // get value.
				gTabCounterPm.globalVal.getSetVal("browserIsBusy", "setF"); // set false.
				gTabCounterPm.globalVal.getSetVal("stateReadyWin", "setV", 0); // set value.
				gTabCounterPm.globalVal.getSetVal("canUpdateOtherWin", "setT"); // set true.
				let delay = 100 * gTabCounterPm.globalVal.getSetVal("restoredWindows", "getV");
				window.setTimeout(gTabCounterPm.updateToolbarItem.bind(gTabCounterPm), delay, "allRestoredWinReady");
					// Session restore may be quite heavy process, especially when being restored several windows/tabs,
					// so let's get out of its way. A little delay here doesn't harm..

				// "domReadyWin": info in "tbcGlobal.jsm".
				gTabCounterPm.globalVal.getSetVal("domReadyWin", "setV", // set value.
					gTabCounterPm.globalVal.getSetVal("restoredWindows", "getV")); // get value.
				gTabCounterPm.globalVal.getSetVal("restoredWindows", "setV", 0); // set value.
				rootReturn = true;
			};
		};
		function addStateReadyWin() {
			let sum = gTabCounterPm.globalVal.getSetVal("stateReadyWin", "getV") + 1;
			gTabCounterPm.globalVal.getSetVal("stateReadyWin", "setV", sum); // set value.
			checkBrowserReadiness();
		};
		if(aEvent.type == "SSWindowStateBusy") {
			window.addEventListener("SSWindowStateReady", function onWinStateReady() {
				window.removeEventListener("SSWindowStateReady", onWinStateReady, false);
				let restoredWinEnum = Services.wm.getEnumerator("navigator:browser");
				while(restoredWinEnum.hasMoreElements()) {
					let restBrWin = restoredWinEnum.getNext();
					if(restBrWin.window) {
						restoredWindows++;
					};
				};
				if((restoredWindows > 1) && (restoredWindows > gTabCounterPm.globalVal.getSetVal("openWindows", "getV"))) { // get value.
				// Important check to make sure that "SSWindowStateBusy" was caused during session restore. This event is fired
				// also in other cases, like "Undo Close Tab", unload tabs, where our global count functionality would be broken
				// after "Undo Close Tab" or "Restore All Tabs". Also, if there is only 1 window, we don't have to proceed furthermore.
					gTabCounterPm.globalVal.getSetVal("browserIsBusy", "setT"); // set true.
					gTabCounterPm.globalVal.getSetVal("canUpdateOtherWin", "setF"); // set false.
					gTabCounterPm.globalVal.getSetVal("restoredWindows", "setV", restoredWindows); // set value.
					addStateReadyWin();
				};
			}, false);
		};
		if(aEvent.type == "SSWindowStateReady") {
			addStateReadyWin();
		};
		return rootReturn;
	},

	// Check and get Preference value.
	getPref: function(name, type) {
		try{
			if(Services.prefs.getPrefType(name) == type) {
				switch(type) {
					case Ci.nsIPrefBranch.PREF_STRING:
						return Services.prefs.getComplexValue(name, Ci.nsISupportsString).data
					case Ci.nsIPrefBranch.PREF_BOOL:
						return Services.prefs.getBoolPref(name);
					case Ci.nsIPrefBranch.PREF_INT:
						return Services.prefs.getIntPref(name);
				};
			};
		} catch(ex) {
			console.log("Tab Counter exception: " + ex);
		};
		return null;
	},

	openOptions: function() {
		var dialog = Services.wm.getMostRecentWindow("TabCounterPm:Options");
		if(dialog) {
			dialog.focus();
			return;
		};
		openDialog("chrome://tabcounterpm/content/tbcOptions.xul", "_blank", "chrome,titlebar,toolbar,centerscreen," + 
			((this.getPref("browser.preferences.instantApply", Ci.nsIPrefBranch.PREF_BOOL))?"dialog=no":"modal"));
	},

	mBundle: null,
	_string: function(aName) {
		if(!this.mBundle) {
			this.mBundle = Services.strings.createBundle("chrome://tabcounterpm/locale/tabcounter.properties");
		};
		return this.mBundle.GetStringFromName(aName);
	}
};
Cu.import("chrome://tabcounter-pm/content/tbcGlobal.jsm", gTabCounterPm);
gTabCounterPm.initialize();
